/**
 * @file    shellmatta_transport.c
 * @brief   transport layer functions of shellmatta
 * @author  Simon Fischer <fischer.simon.1991@gmail.com>
 */

#include "shellmatta_transport.h"
#include "shellmatta_crc.h"
#include <string.h>
#include <stdio.h>

/* init global variables */
uint32_t sentPayloadBytes = 0;

uint8_t protocolVersion                     = 0u;
shellmatta_transport_packet_t packetType    = 0u;
uint8_t payloadLength                       = 0u;
uint8_t source                              = 0u;
uint8_t destination                         = 0u;
uint32_t crc32                              = 0u;
uint8_t payloadCounter                      = 0u;
uint8_t crcCounter                          = 0u;
uint8_t packetSequenceCounter_h2s           = 0u;
uint8_t packetSequenceCounter_s2h           = 0u;
shellmatta_transport_layer_t transportLayerInst = {
    0,
    0,
    STATE_GET_SOH,
    false,
    false,
    false,
    NULL
};
char payloadBuffer[SHELLMATTA_PAYLOAD_MAXLENGTH + 1];

shellmatta_retCode_t shellmatta_init_transport_inst()
{
    transportLayerInst.h2s_sequenceCnt = 0;
    transportLayerInst.s2h_sequenceCnt = 0;
    transportLayerInst.state = STATE_GET_SOH;
    transportLayerInst.mandatory = false;
    transportLayerInst.active = false;
    transportLayerInst.continueStep = false;
    memset(payloadBuffer, 0, SHELLMATTA_PAYLOAD_MAXLENGTH + 1);

    return SHELLMATTA_OK;
}

shellmatta_retCode_t shellmatta_reset_transport()
{
    transportLayerInst.state = STATE_GET_SOH;
    protocolVersion = 0u;
    packetType      = 0u;
    payloadLength   = 0u;
    source          = 0u;
    destination     = 0u;
    crc32           = 0u;
    payloadCounter  = 0u;
    crcCounter      = 0u;
    transportLayerInst.active = false;
    memset(payloadBuffer, 0, SHELLMATTA_PAYLOAD_MAXLENGTH + 1);

    return SHELLMATTA_OK;
}

shellmatta_retCode_t shellmatta_handle_transport_fsm(char *data)
{
    switch (transportLayerInst.state)
    {
    case STATE_GET_SOH:
        if (SHELLMATTA_START_OF_HEADER == *data)
        {
            transportLayerInst.state = STATE_GET_PROTOCOL_VERSION;
            transportLayerInst.active = true;
        }
        else
        {
            transportLayerInst.state = STATE_MANUAL_INPUT;
            transportLayerInst.active = false;
        }
        break;

    case STATE_MANUAL_INPUT:
        break;

    case STATE_GET_PROTOCOL_VERSION:
        transportLayerInst.state = STATE_GET_PACKET_TYPE;
        break;

    case STATE_GET_PACKET_TYPE:
        transportLayerInst.state = STATE_GET_PAYLOAD_LENGTH;
        break;

    case STATE_GET_PAYLOAD_LENGTH:
        transportLayerInst.state = STATE_GET_SOURCE;
        break;

    case STATE_GET_SOURCE:
        transportLayerInst.state = STATE_GET_DESTINATION;
        break;

    case STATE_GET_DESTINATION:
        transportLayerInst.state = STATE_GET_H2S_SEQUENCE_CNT;
        break;

    case STATE_GET_H2S_SEQUENCE_CNT:
        transportLayerInst.state = STATE_GET_S2H_SEQUENCE_CNT;
        break;

    case STATE_GET_S2H_SEQUENCE_CNT:
        if (packetType == PACKET_SEQ_CNT_REQUEST)
        {
            transportLayerInst.state = STATE_GET_CRC;
        }
        else
        {
            transportLayerInst.state = STATE_GET_PAYLOAD;
        }
        break;

    case STATE_GET_PAYLOAD:
        if (    (payloadLength <= payloadCounter)
            &&  (true == transportLayerInst.active))
        {
            transportLayerInst.state = STATE_GET_CRC;
        }
        break;

    case STATE_GET_CRC:
        if (SHELLMATTA_LENGTH_CRC <= crcCounter)
        {
            
            char crcdata[SHELLMATTA_PAYLOAD_MAXLENGTH + 1 + SHELLMATTA_HEADER_LENGTH];
            memset(crcdata, 0, SHELLMATTA_PAYLOAD_MAXLENGTH + 1 + SHELLMATTA_HEADER_LENGTH);
            crcdata[0] = SHELLMATTA_START_OF_HEADER;
            crcdata[1] = protocolVersion;
            crcdata[2] = packetType;
            crcdata[3] = payloadLength;
            crcdata[4] = source;
            crcdata[5] = destination;
            crcdata[6] = packetSequenceCounter_h2s;
            crcdata[7] = packetSequenceCounter_s2h;
            strncat(&crcdata[8], payloadBuffer, payloadLength);

            uint32_t refCrc;

            refCrc = crc32Calc(crcdata, SHELLMATTA_HEADER_LENGTH + payloadLength);

            if (crc32 == refCrc)
            {
                crcCounter = 0;
                crc32 = 0;
                transportLayerInst.state = STATE_GET_SOH;

                /* crc is correct */
                transportLayerInst.h2s_sequenceCnt++;

                switch (packetType)
                {
                case PACKET_DATA:
                    transportLayerInst.state = STATE_PROCESS_PAYLOAD;
                    break;
                
                case PACKET_SEQ_CNT_REQUEST:
                    /* change packet type to response */
                    packetType = PACKET_SEQ_CNT_RESPOND;
                    /* send out packet with no payload */
                    shellmatta_write_transport("", 0);
                    if (transportLayerInst.h2s_sequenceCnt != packetSequenceCounter_h2s)
                    {
                        crc32 = 0;
                    }
                    break;

                case PACKET_SEQ_CNT_RESPOND:
                    break;
                
                case PACKET_MAX_BUFFERSIZE_REQUEST:
                    break;

                case PACKET_MAX_BUFFERSIZE_RESPOND:
                    break;
                
                case PACKET_SEARCH_DEVICE_REQUEST:
                    break;
                
                case PACKET_SEARCH_DEVICE_RESPOND:
                    break;

                case PACKET_SET_ADDRESS_REQUEST:
                    break;

                case PACKET_SET_ADDRESS_RESPOND:
                    break;

                /* wrong packet type */
                default:
                    /* undo sequence counter increment */
                    transportLayerInst.h2s_sequenceCnt--;
                    break;
                }
                break;
            }
            else
            {
                // TODO: improve this
                crcCounter = 0;
                crc32 = 0;
                transportLayerInst.state = STATE_GET_SOH;

                /* crc is incorrect */
                return SHELLMATTA_ERROR;
            }
        }
        break;

    case STATE_PROCESS_PAYLOAD:
        transportLayerInst.state = STATE_GET_SOH;
        protocolVersion = 0u;
        packetType      = 0u;
        payloadLength   = 0u;
        source          = 0u;
        destination     = 0u;
        crc32           = 0u;
        payloadCounter  = 0u;
        crcCounter      = 0u;
        transportLayerInst.active = false;
        memset(payloadBuffer, 0, SHELLMATTA_PAYLOAD_MAXLENGTH + 1);
        break;
    
    default:
        break;
    }

    return SHELLMATTA_OK;
}

shellmatta_retCode_t shellmatta_write_transport(const char* data, uint32_t length)
{
    /* create buffer for data + header + crc with maximum size */
    uint8_t outputBuffer[   SHELLMATTA_HEADER_LENGTH +          /* length of header */
                            SHELLMATTA_PAYLOAD_MAXLENGTH + 1 +  /* max length of payload */
                            SHELLMATTA_LENGTH_CRC];             /* length of crc */
    
    uint32_t outPayloadLength = length + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_LENGTH_CRC;

    /* fill buffer with header data */
    outputBuffer[0] = SHELLMATTA_START_OF_HEADER;           /* start of header */
    outputBuffer[1] = PROTOCOL_VERSION;                     /* protocol version */
    outputBuffer[2] = packetType;                           /* packet type */
    outputBuffer[3] = length;                               /* payload length */
    outputBuffer[4] = 0x00u;                                /* source */
    outputBuffer[5] = 0x00u;                                /* destination */
    outputBuffer[6] = transportLayerInst.h2s_sequenceCnt;   /* sequence counter host to shellmatta */
    outputBuffer[7] = ++transportLayerInst.s2h_sequenceCnt; /* sequence counter shellmatta to host */

    /* response of sequence counters is only packet type without payload */
    if (packetType != PACKET_SEQ_CNT_RESPOND)
    {
        memcpy(&outputBuffer[8], data, length);
    }

    uint32_t outCrc = crc32Calc((char*) outputBuffer, SHELLMATTA_HEADER_LENGTH + length);

    /* append crc to end of payload */
    outputBuffer[length + SHELLMATTA_HEADER_LENGTH + 0] = (uint8_t)(outCrc >> 0);
    outputBuffer[length + SHELLMATTA_HEADER_LENGTH + 1] = (uint8_t)(outCrc >> 8);
    outputBuffer[length + SHELLMATTA_HEADER_LENGTH + 2] = (uint8_t)(outCrc >> 16);
    outputBuffer[length + SHELLMATTA_HEADER_LENGTH + 3] = (uint8_t)(outCrc >> 24);

    /* use original write function to send full buffer */
    shellmatta_retCode_t ret = transportLayerInst.originalWrite((char*) outputBuffer, outPayloadLength);
    return ret;
}
